﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Threading.Tasks.Dataflow;
using Models.Events;
using Models.Exceptions;
using Models.Rules;

namespace Models.Game
{
    /// <summary>
    /// The Game class represents a game session in the application.
    /// It contains all the necessary properties and methods to manage a game, including the game loop, dice rolling, and use of the game rules.
    /// </summary>
    public class Game
    {
        private bool _isRunning;
        public Player CurrentPlayer { get; private set; }

        public Map UsedMap { get; private set; }

        private Dice Dice1 { get; }
        private Dice Dice2 { get; }

        private int Turn { get; set; }

        public Rules.Rules GameRules { get; }

        
        // == Events ==
        public event EventHandler<GameStartedEventArgs> GameStarted;
        public event EventHandler<GameEndedEventArgs> GameEnded;
        public event EventHandler BoardUpdated;
        public event EventHandler<DiceRolledEventArgs> DiceRolled;
        public event EventHandler<OperationChosenEventArgs> OperationChosen;
        public event EventHandler<CellChosenEventArgs> CellChosen;
     
        
        /// <summary>
        /// Initializes a new instance of the <see cref="Game"/> class.
        /// </summary>
        /// <param name="player">The player who play the game.</param>
        /// <param name="map">The map to be used in the game.</param>
        public Game(Player player, Map map)
        {
            _isRunning = false;
            UsedMap = map;
            CurrentPlayer = player;
            Dice1 = new Dice();
            Dice2 = new Dice(1);
            Turn = 1;
            GameRules = new Rules.Rules();
        }

        /// <summary>
        /// Rolls all the dice.
        /// </summary>
        private void RollAllDice()
        {
            Dice1.Roll();
            Dice2.Roll();
            DiceRolled?.Invoke(this, new DiceRolledEventArgs(Dice1.Value, Dice2.Value));
        }
        
        /// <summary>
        /// Marks an operation as checked in the operation grid of the game.
        /// </summary>
        /// <param name="operation"></param>
        private void MarkOperationAsChecked(Operation operation)
        {
            int operationIndex = (int)operation;
            int operationsPerType = 4;  // Chaque type d'opération peut être fait 4 fois

            for (int i = operationIndex * operationsPerType; i < (operationIndex + 1) * operationsPerType; i++)
            {
                if (!UsedMap.OperationGrid[i].IsChecked)
                {
                    UsedMap.OperationGrid[i].Check();
                    break;
                }
            }
        }

        /// <summary>
        /// Performs an operation on the values of two dice based on the provided operation.
        /// </summary>
        /// <param name="o">The operation to perform. This can be LOWER, HIGHER, SUBTRACTION, ADDITION, or MULTIPLICATION.</param>
        /// <returns>
        /// The result of the operation. If the operation is LOWER or HIGHER, it returns the lower or higher value of the two dice respectively.
        /// If the operation is SUBTRACTION, it returns the difference between the higher and lower value of the two dice.
        /// If the operation is ADDITION, it returns the sum of the values of the two dice.
        /// If the operation is MULTIPLICATION, it returns the product of the values of the two dice.
        /// If the operation is not one of the operations, it throws an ArgumentOutOfRangeException.
        /// </returns>
        private int ResultOperation(Operation o)
        {
            int result = o switch
            {
                Operation.LOWER => Dice1.IsLower(Dice2) ? Dice1.Value : Dice2.Value,
                Operation.HIGHER => Dice1.IsLower(Dice2) ? Dice2.Value : Dice1.Value,
                Operation.SUBTRACTION => Dice1.IsLower(Dice2) ? Dice2.Value - Dice1.Value : Dice1.Value - Dice2.Value,
                Operation.ADDITION => Dice2.Value + Dice1.Value,
                Operation.MULTIPLICATION => Dice2.Value * Dice1.Value,
                _ => throw new ArgumentOutOfRangeException()
            };

            MarkOperationAsChecked(o);
            return result;
        }


        /// <summary>
        /// Places the result of a dice operation into a chosen cell on the game board.
        /// The result can be placed in the chosen cell if it's the first turn or if the chosen cell is valid according to the game rules.
        /// </summary>
        /// <param name="playerChoice">The cell chosen by the player to place the result.</param>
        /// <param name="result">The result of the dice operation to be placed in the cell.</param>
        private void PlaceResult(Cell playerChoice, int result)
        {
            if (Turn == 1 || GameRules.NearCellIsValid(playerChoice, UsedMap.Boards))
            {
                playerChoice.Value = result;
                BoardUpdated?.Invoke(this, EventArgs.Empty);
            }
        }

        /// <summary>
        /// Add the choosen cell to a rope path if it's possible.
        /// </summary>
        /// <param name="playerChoice"></param>
        /// <param name="adjacentes"></param>
        private void AddToRopePath(Cell playerChoice,List<Cell> adjacentes)
        {
            int index =0;

            foreach (var cells in adjacentes.Where(cells => cells.Value - playerChoice.Value == 1 || cells.Value - playerChoice.Value == -1))
            {
                // Le cas si il n'existe aucun chemin de corde
                if (UsedMap.RopePaths.Count == 0)
                {
                    // Creer un nouveau chemin de corde avec la cellule choisi par le joueur et celle adjacente
                    UsedMap.RopePaths.Add(new List<Cell> {playerChoice, cells});
                }


                // A modifier dans le cas ou il est possible de fusionner deux chemins de corde
                if (GameRules.IsInRopePaths(playerChoice, UsedMap.RopePaths, index)) break;

                // Le cas si il existe des chemins de corde

                // Est-ce que la cellule adjacentes fait parti d'un chemin de corde
                if (!GameRules.IsInRopePaths(cells,UsedMap.RopePaths,index))
                {
                    UsedMap.RopePaths.Add(new List<Cell> { playerChoice, cells });
                    continue;
                }

                if (!GameRules.AsValue(playerChoice,UsedMap.RopePaths,index))
                {
                    UsedMap.RopePaths[index].Add(playerChoice);
                }
            }
        }
        
        /// <summary>
        /// Initializes the game.
        /// </summary>
        public void InitializeGame()
        {
            _isRunning = true;
            GameStarted?.Invoke(this, new GameStartedEventArgs(CurrentPlayer));
            GameLoop();
        }

        /// <summary>
        /// Ends the game.
        /// </summary>
        private void EndGame()
        {
            _isRunning = false;
            GameEnded?.Invoke(this, new GameEndedEventArgs(CurrentPlayer));
        }
        
        /// <summary>
        /// The main game loop that runs while the game is active.
        /// </summary>
        private void GameLoop()
        {
            while (_isRunning)
            {
                if (Turn == 20)
                {
                    EndGame();
                    break;
                }
                
                RollAllDice();

                Turn++;
            }
        }
        
        /// <summary>
        /// Handles the player's choice of an operation based on the dice values.s
        /// </summary>
        /// <param name="operation">The operation chosen by the player.</param>
        public void HandlePlayerOperation(Operation operation)
        {
            int result = ResultOperation(operation);
            OperationChosen?.Invoke(this, new OperationChosenEventArgs(operation, result));
        }

        /// <summary>
        /// Handles the player's choice of a cell on the game board.
        /// </summary>
        /// <param name="cell"></param>
        /// <param name="result"></param>
        /// <exception cref="InvalidCellCoordinatesException"></exception>
        /// <exception cref="InvalidCellException"></exception>
        public void HandlePlayerChoice(Cell cell, int result)
        {
            if (cell.X < 0 || cell.X >= UsedMap.Boards.Count / 6 || cell.Y < 0 || cell.Y >= 6)
            {
                throw new InvalidCellCoordinatesException("Invalid cell coordinates. Please choose again.");
            }

            if (!GameRules.IsCellValid(cell, UsedMap.Boards))
            {
                throw new InvalidCellException("Cell is not valid. Please choose again.");
            }

            PlaceResult(cell, result);
            GameRules.IsZoneValidAndAddToZones(cell, UsedMap);
            AddToRopePath(cell, GameRules.EveryAdjacentCells(cell, UsedMap.Boards));
            CellChosen?.Invoke(this, new CellChosenEventArgs(cell, result));
            BoardUpdated?.Invoke(this, EventArgs.Empty);
        }
    }
}
